@@ -61,7 +61,7 @@ CONFIG_FILE = '~/.bbrf/config.json'
 REGEX_DOMAIN = re.compile('^(?:[a-z0-9_](?:[a-z0-9-_]{0,61}[a-z0-9])?\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$')
 # regex to match IP addresses and CIDR ranges - thanks https://www.regextester.com/93987
 REGEX_IP = re.compile('^([0-9]{1,3}\.){3}[0-9]{1,3}(\/([0-9]|[1-2][0-9]|3[0-2]))?$')
-VERSION = '1.1.3'
+VERSION = '1.1.4'
 
 class BBRFClient:
     config = {}
@@ -149,19 +149,6 @@ class BBRFClient:
             raise Exception('This program does not exist.')
         self.config['program'] = pro
     
-    '''
-    Check whether a domain matches a scope
-    '''
-    def matches_scope(self, domain, scope):
-        for s in scope:
-            # literal match
-            if s == domain:
-                return True
-            # x.example.com matches *.example.com
-            if s.startswith('*.') and domain.endswith('.'+s[2:]):
-                return True
-        return False
-    
     '''
     Check whether an IP belongs to a CIDR range
     '''
@@ -284,15 +271,20 @@ class BBRFClient:
             # Avoid adding blacklisted domains or
             # domains that resolve to blacklisted ips
             if blacklisted_ip or domain in blacklist:
+                self.debug('blacklisted: '+domain)
                 continue
             # It must match the format of a domain name
             if not REGEX_DOMAIN.match(domain):
+                self.debug('REGEX_DOMAIN failed: '+domain)
                 continue
             # It may not be explicitly outscoped
             if self.matches_scope(domain, outscope):
+                self.debug('outscope: '+domain)
                 continue
-            # It must match the in scope
-            if not self.matches_scope(domain, inscope):
+            # It must match the in scope, except if we're trying to @INFER the program later,
+            # which means we cannot verify the scope here
+            if not self.get_program() == '@INFER' and not self.matches_scope(domain, inscope):
+                self.debug('Not inscope: '+domain)
                 continue
                 
             # Add the ips if we already parsed other ips for this domain,
@@ -301,7 +293,7 @@ class BBRFClient:
                 add_domains[domain].extend(ips)
             else:
                 add_domains[domain] = ips
-        
+                
         # add all new scope at once to drastically reduce runtime of large input
         if len(add_inscope) > 0:
             self.add_inscope(add_inscope, passalong={'doc':copy, 'inscope':inscope, 'outscope':outscope})
@@ -526,11 +518,11 @@ class BBRFClient:
                 print("Illegal hostname:",hostname)
                 continue
             # It may not be explicitly outscoped
-            if self.matches_scope(hostname, outscope):
+            if not self.get_program() == '@INFER' and self.matches_scope(hostname, outscope):
                 print("skipping outscoped hostname:",hostname)
                 continue
             # It must match the in scope
-            if not self.matches_scope(hostname, inscope):
+            if not self.get_program() == '@INFER' and not self.matches_scope(hostname, inscope):
                 print("skipping not inscope hostname:",hostname)
                 continue
             
@@ -589,6 +581,13 @@ class BBRFClient:
         remove = {url.split(" ")[0]: {'_deleted': True} for url in urls}
         _ = self.api.update_documents('url', remove)
         
+    '''
+    Remove Services
+    '''
+    def remove_services(self, services):
+        remove = {service: {'_deleted': True} for service in services}
+        _ = self.api.update_documents('service', remove)
+        
         
     '''
     Store services (i.e. open ports) of IP addresses
@@ -811,6 +810,20 @@ class BBRFClient:
             program_name = False
         return self.api.get_tags(tagname, program_name)
         
+    '''
+    Check whether a domain matches a scope
+    '''
+    @staticmethod
+    def matches_scope(domain, scope):
+        for s in scope:
+            # literal match
+            if s == domain:
+                return True
+            # x.example.com matches *.example.com
+            if s.startswith('*.') and domain.endswith('.'+s[2:]):
+                return True
+        return False    
+    
     def debug(self, msg):
         if 'debug' in self.config and self.config['debug']:
             print('[DEBUG] '+msg)
@@ -955,6 +968,12 @@ class BBRFClient:
                     return self.add_services(self.arguments['<service>'])
                 if self.arguments['-']:
                     return self.add_services([u.rstrip() for u in sys.stdin.read().split('\n')])
+            if self.arguments['remove']:
+                if self.arguments['<service>']:
+                    return self.remove_services(self.arguments['<service>'])
+                if self.arguments['-']:
+                    return self.remove_services([s.rstrip() for s in sys.stdin.read().split('\n')])
+                
             
         if self.arguments['services']:
             if self.arguments['where']:
@@ -1038,6 +1057,9 @@ def main():
                 print(result)
     except Exception as e:
         print('[ERROR] '+str(e))
+        if 'debug' in self.config and self.config['debug']:
+            import traceback
+            traceback.print_exc()
             
 if __name__ == '__main__':
     main()
\ No newline at end of file
